@@ -151,6 +151,8 @@ Field types
    :members: encoder, decoder, default, default_enc
    
 .. autoclass:: Bytes
+
+.. autoclass:: String
    
 .. autoclass:: Pickle
    
@@ -167,7 +169,7 @@ Utility functions
 
 .. autofunc:: full_header
 '''
-__all__ = ['BinPack', 'Field', 'Bytes', 'Pickle', 'Compressed',
+__all__ = ['BinPack', 'Field', 'Bytes', 'String', 'Pickle', 'Compressed',
            'raw_unpack', 'header', 'version', 'full_header']
 
 import collections
@@ -176,13 +178,111 @@ try:
 except ImportError:
     from StringIO import StringIO
 
-class voiddict(collections.MutableMapping):
+#===============================================================================
+# Auxiliary types
+#===============================================================================
+class lazydict(collections.MutableMapping):
+    '''A lazy dictionary created from an iterator over (key, values).
+    
+    `lazydict` objects are initialized from a possibly infinite iterator over 
+    (key, value) pairs and, optionally, a function `goodkeys` that marks if a 
+    key will appear in the iterator or not. `goodkeys` is either a function 
+    that return True for keys that appear in the iterator and False otherwise, 
+    or a container type such as dict, list, tuple or set holding the valid keys.
+    '''
+
+    def __init__(self, items, goodkeys=None):
+        self._items = iter(items)
+        self._consumed = set()
+        self._dict = collections.OrderedDict()
+        if goodkeys is None:
+            self._isgood = lambda x: True
+        elif callable(goodkeys):
+            self._isgood = goodkeys
+        else:
+            if isinstance(goodkeys, (list, tuple, dict)):
+                self._isgood = set(goodkeys).__contains__
+            else:
+                self._isgood = goodkeys.__contains__
+
+    def consume(self, *args):
+        '''Consume internal items iterator up to the given key'''
+
+        if len(args) == 1:
+            stop_at = args[0]
+        elif not args:
+            stop_at = object()
+        else:
+            raise TypeError('accepts 1 or 2 arguments, %s given' % (len(args) + 1))
+
+        if (stop_at in self._consumed) or (self._items is None) or (not self._isgood(stop_at)):
+            return
+        else:
+            while True:
+                try:
+                    k, v = next(self._items)
+                except StopIteration:
+                    break
+                else:
+                    if k not in self._consumed:
+                        self._dict[k] = v
+                        self._consumed.add(k)
+                        if k == stop_at:
+                            break
+                    else:
+                        raise ValueError('items have a duplicated key: %s' % k)
+            else:
+                self._items = None
+
+    def __delitem__(self, key):
+        if self._items is None or key in self._consumed:
+            del self._dict[key]
+        else:
+            self.consume
+        if key in self._dict:
+            del self._dict[key]
+        else:
+            self.consume(key)
+            if key in self._dict:
+                del self._dict[key]
+
+    def __getitem__(self, key):
+        try:
+            return self._dict[key]
+        except KeyError:
+            self.consume(key)
+            return self._dict[key]
+
+    def __iter__(self):
+        for k in self._dict:
+            yield k
+        for k, v in self._items:
+            self._dict[k] = v
+            self._consumed.add(k)
+            yield k
+        self._items = None
+
+    def __len__(self):
+        self.consume()
+        return len(self._dict)
+
+    def __setitem__(self, key, value):
+        self.consume(key)
+        self._dict[key] = value
+
+class nulldict(collections.MutableMapping):
+    '''All keys entered in a `nulldict` are sent to trash (or maybe to /dev/null)'''
+
     def __delitem__(self, key):
         raise KeyError(key)
+
     __getitem__ = __setitem__ = __delitem__
 
     def __iter__(self):
-        return iter(())
+        if True:
+            return
+        else:
+            yield
 
     def __len__(self):
         return 0
@@ -215,11 +315,11 @@ class BinPack(collections.Mapping):
     class __metaclass__(type(collections.Mapping)):
         def __new__(cls, name, bases, namespace):
 
-            # Initializes the BinPack class
+            # Creates the BinPack class
             if BinPack is None:
                 return type(collections.Mapping).__new__(cls, name, bases, namespace)
 
-            # Initializes a BinPack sub-class
+            # Creates a BinPack sub-class
             else:
                 # Check bases
                 if len(bases) > 1:
@@ -249,10 +349,8 @@ class BinPack(collections.Mapping):
             raise TypeError(msg)
 
     def __init__(self, F, mode='r', close=False, atomic=False, keep_items=True):
-        self._processed = set()
-        self._raw = {}
-        self._items = {} if keep_items else voiddict()
-        self._ok = []
+        self._undefined = object()
+        self._items = {} if keep_items else nulldict()
         self._sep = '\xff' if self.get_header()[0] == '\xff' else '\n'
         self._fields = self._cls_dict['fields']
         self._close = close
@@ -279,20 +377,32 @@ class BinPack(collections.Mapping):
 
         # Reads/writes the header/version to file
         if self.mode == 'r':
-            self._data = raw_unpack(self.file)
-            header = next(self._data)
-            version = next(self._data)
+            init_pos = self.file.tell()
+            try:
+                data = raw_unpack(self.file)
+                header = next(data)
+                version = next(data)
+                del data
+            finally:
+                self.file.seek(init_pos)
 
             if header != self.get_header():
                 raise ValueError("bad header: %s" % repr(header))
             if version != self.get_version():
                 raise ValueError("bad version: %s" % repr(version))
 
-            self._fields_it = iter(self._fields)
-            self._data = ((next(self._fields_it), data) for data in self._data)
+            self._raw = lazydict(self._read_data())
         else:
+            self._raw = {}
             self.file.write(self._cls_dict['full_header'])
 
+    def _read_data(self):
+        data = raw_unpack(self.file)
+        next(data)
+        next(data)
+        for f in self:
+            yield f, next(data)
+
     #===========================================================================
     # Context Manager/File
     #===========================================================================
@@ -312,14 +422,14 @@ class BinPack(collections.Mapping):
 
         try:
             if self.mode == 'w':
-                keys = set(self) - self._processed
-                fields = self._fields
-                for key in keys:
-                    default = fields[key].default()
-                    if default is not None:
-                        self[key] = default
-                    else:
-                        raise ValueError("missing data for field '%s'" % key)
+                for k in self:
+                    if k not in self._raw:
+                        default = self._fields[k].default()
+                        if default is None:
+                            raise ValueError("missing data for field '%s'" % k)
+                        else:
+                            self._raw[k] = default
+                self._write_defined()
 
                 if self._file_after is not None:
                     data = self.file.getvalue()
@@ -333,14 +443,6 @@ class BinPack(collections.Mapping):
         if self.closed:
             raise ValueError('I/O operation on closed file')
 
-    def _write(self, data):
-        self._assure_opened()
-        self.file.write(bytes7num(len(data)))
-        self.file.write(data)
-
-    #===========================================================================
-    # Properties
-    #===========================================================================
     @property
     def _cls_dict(self):
         return type(self)._cls_dict_
@@ -350,45 +452,34 @@ class BinPack(collections.Mapping):
     #===========================================================================
     def __getitem__(self, key):
         self._assure_opened()
-        if key not in self._fields:
-            raise KeyError('invalid key: %s' % key)
-        elif key in self._items:
+        if key in self._items:
             return self._items[key]
-        elif key in self._processed:
-            raise KeyError("already retrieved: %s" % key)
+        else:
+            if key not in self._fields:
+                raise KeyError('invalid key: %s' % key)
+
+            # In read mode, it tries to fetch from the file
+            if self.mode == 'r':
+                data = self._raw[key]
+                self._raw[key] = self._undefined
+
+                if data is self._undefined:
+                    raise KeyError('key was consumed: %s' % key)
 
-        # In read mode, it tries to fetch from the file
-        if self.mode == 'r':
-            # Check if data is in raw
-            try:
-                data = self._raw.pop(key)
-            except KeyError:
-                pass
-            else:
                 decoder = self._fields[key].decoder
                 value = decoder(data)
-                self._processed.add(key)
                 self._items[key] = value
                 return value
 
-            # Tries to fetch data from file
-            for k, data in self._data:
-                self._raw[k] = data
-                if k == key:
-                    return self[key]
-
-            raise RuntimeError('should never reach here\n\tkey: %s\n\tdata: %s' % (key, self._raw.keys()))
-
-        # In write mode, it checks for default values
-        else:
-            try:
-                return self._raw[key]
-            except KeyError:
-                default = self._fields[key].default()
-                if default is not None:
-                    return default
-                else:
-                    raise KeyError(key)
+            # In write mode, it checks for default values
+            else:
+                try:
+                    return self._raw[key]
+                except KeyError:
+                    default = self._fields[key].default()
+                    if default is not None:
+                        return default
+                    raise
 
     def __iter__(self):
         return iter(self._fields)
@@ -396,29 +487,39 @@ class BinPack(collections.Mapping):
     def __len__(self):
         return len(self._fields)
 
-    #===========================================================================
-    # Magic methods
-    #===========================================================================
     def __setitem__(self, key, value):
         self._assure_opened()
         if self.mode == 'r':
             raise ValueError('cannot set items in read mode')
-
-        if key in self._processed:
+        elif key not in self._fields:
+            raise KeyError('invalid key: %s' % key)
+        elif key in self._raw:
             raise ValueError('key already saved to file')
+
         self._items[key] = self._raw[key] = value
-        keys = list(self)
-        for k in keys[len(self._ok):]:
+        self._write_defined()
+
+    def _write_defined(self):
+        self._assure_opened()
+        for k in self:
             try:
-                value = self._raw.pop(k)
+                value = self._raw[k]
             except KeyError:
                 break
             else:
-                encoder = self._fields[k].encoder
-                data = encoder(value)
-                self._write(data)
-                self._ok.append(k)
-                self._processed.add(k)
+                if value is self._undefined:
+                    continue
+                else:
+                    encoder = self._fields[k].encoder
+                    try:
+                        data = encoder(value)
+                    except Exception as ex:
+                        print 'error found writing key: %s' % k
+                        raise
+
+                    self.file.write(bytes7num(len(data)))
+                    self.file.write(data)
+                    self._raw[k] = self._undefined
 
     #===========================================================================
     # Subtypes creation
@@ -437,12 +538,6 @@ class BinPack(collections.Mapping):
             NewType.set_version(version)
         return NewType
 
-#TODO: implement versioning
-#    @classmethod
-#    def new_version(cls, version=None):
-#        cls._assure_locked()
-#        raise NotImplementedError
-
     #===========================================================================
     # Locking mechanism
     #===========================================================================
@@ -625,6 +720,27 @@ class Bytes(Field):
     def decoder(self, data):
         return data
 
+class String(Field):
+    '''Data is a unicode string saved using the specified encoding.'''
+
+    def __init__(self, default=None, encoding='utf8'):
+        super(String, self).__init__(default)
+        self.encoding = encoding
+
+    def encoder(self, obj):
+        if isinstance(obj, unicode):
+            return obj.encode(self.encoding)
+        elif isinstance(obj, basestring):
+            return unicode(obj, self.encoding).encode(self.encoding)
+        else:
+            raise TypeError('expected string, got %s' % type(obj))
+
+    def decoder(self, data):
+        return data.decode(self.encoding)
+
+
+
+
 class Serializer(Field):
     def serializers(self):
         '''It must return a tuple of (encoder, decoder) functions. The arguments
@@ -689,6 +805,15 @@ class Pickle(Serializer):
 class JSON(Serializer):
     '''Serialize a JSON structure.
     
+    The JSON serializer accepts the same arguments as the `json.dumps` and 
+    `json.loads` functions with two small differences:
+    
+        * The ``cls`` parameter in each dumps/loads function should now be a 
+          tuple (JSONEncoder or None, JSONDecoder or None).
+          
+        * The `default` parameter of json_default is now named `json_default`
+          in order to prevent clashing with the existing `default` parameter.  
+    
     Examples
     --------
     
@@ -702,7 +827,24 @@ class JSON(Serializer):
     '...{"foo": "bar", "spam": ["ham", "eggs"]}'
     '''
 
-    def serializers(self):
+    def __init__(self, default=None,
+                 # common
+                 encoding='utf8', cls=(None, None), use_decimal=False,
+
+                 # encoding options
+                 skipkeys=False, check_circular=True, ensure_ascii=True,
+                 allow_nan=True, indent=None, separators=None, json_default=None,
+
+                 # decoding options
+                 object_hook=None, object_pairs_hook=None,
+                 parse_float=None, parse_int=None, parse_constant=None):
+
+        kwds = dict(locals())
+        kwds.pop('default')
+        kwds.pop('self')
+        super(JSON, self).__init__(default, **kwds)
+
+    def serializers(self, **kwds):
         '''Loads dumps and loads functions from the simplejson or json modules'''
 
         try:
@@ -710,7 +852,25 @@ class JSON(Serializer):
         except ImportError:
             import json
 
-        return (json.dumps, json.loads)
+        kwds.setdefault('default', kwds.pop('json_default', None))
+        dump_keys = ['encoding', 'cls', 'use_decimal', 'skipkeys',
+                     'check_circular', 'ensure_ascii', 'allow_nan', 'indent',
+                     'separators', 'default']
+        load_keys = ['encoding', 'cls', 'use_decimal', 'object_hook',
+                     'object_pairs_hook', 'parse_float', 'parse_int',
+                     'parse_constant']
+
+        dumps_kwds = { k: kwds[k] for k in dump_keys }
+        loads_kwds = { k: kwds[k] for k in load_keys }
+        dumps_kwds['cls'], loads_kwds['cls'] = kwds['cls']
+
+        def dumps(obj):
+            return json.dumps(obj, **dumps_kwds)
+
+        def loads(data):
+            return json.loads(data, **loads_kwds)
+
+        return (dumps, loads)
 
 class Compressed(Serializer):
     '''Compressed can be applied to any field in order to compress an 
@@ -835,10 +995,7 @@ def raw_unpack(F, sep='\xff'):
     while n is not None:
         n = read_num(F)
         if n is not None:
-            yield F.read(n)
-
-    remaining = F.read()
-    assert not remaining, 'some data was not read: %s' % remaining
+            yield F.read(n) if n else ''
 
 def header(F):
     '''Returns the header of a file.'''
@@ -879,4 +1036,4 @@ if __name__ == '__main__':
 #        print data['data']
 
     import doctest
-#    doctest.testmod(optionflags=doctest.REPORT_ONLY_FIRST_FAILURE, verbose=0)
+    doctest.testmod(optionflags=doctest.REPORT_ONLY_FIRST_FAILURE, verbose=0)
